package com.sgr_b2.compass.activities;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.Intent;
import android.content.SharedPreferences;
import android.location.Location;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewTreeObserver;
import android.view.ViewTreeObserver.OnGlobalLayoutListener;
import android.view.WindowManager;
import android.view.animation.RotateAnimation;
import android.widget.ImageView;
import android.widget.TextView;

import com.sgr_b2.compass.Config;
import com.sgr_b2.compass.R;
import com.sgr_b2.compass.db.POI;
import com.sgr_b2.compass.jni.Screen;
import com.sgr_b2.compass.jni.Utils;
import com.sgr_b2.compass.navigation.Monitor;
import com.sgr_b2.compass.phone.DeviceWrapper;
import com.sgr_b2.compass.themes.Theme;
import com.sgr_b2.compass.themes.ThemesFactory;
import com.sgr_b2.compass.ui.DontShowAgainOKListener;
import com.sgr_b2.compass.ui.UnitsFormatting;


public class MainActivity extends Activity implements OnGlobalLayoutListener {

	private static final int REQUEST_BOOKMARK_RESULT = 1001;

	private Location location = null;
	private POI target = null;

	private int graphics_state = -1; // used for graphics update

	private Monitor monitor = null;
	private Screen screen = new Screen(Config.COMPASS_ANIMATION_THRESHOLD);
	private DeviceWrapper device_wrapper = null;
	private Menu menu = null; // this is here to manually update menu items on Android 1.x-2.x

	private Theme theme = null;

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		this.device_wrapper = new DeviceWrapper(this);

		if (!this.device_wrapper.isMagneticSensorAvailable()) {
			maybeShowCompassSensorsAlert();
		}

		setContentView(R.layout.layout_main);
	}

	private void maybeStartDistressCallService() {
		final SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);
		boolean run_service = shared_pref.getBoolean(SettingsActivity.KEY_PREF_DISTRESS_RUN_SERVICE, true);

		if (run_service) {
			ActivityUtils.startDistressCallService(this);
		}
		else {
			ActivityUtils.stopDistressCallService(this);
		}
	}

	private void maybeShowCompassSensorsAlert() {
		final SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);
		boolean show_sensors_alert = shared_pref.getBoolean(SettingsActivity.KEY_PREF_SHOW_SENSORS_ALERT, true);

		if (!show_sensors_alert) {
			return;
		}

		LayoutInflater layout_inflater = LayoutInflater.from(this);
		final View alert_view = layout_inflater.inflate(R.layout.layout_showonce_alert, null);

		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setTitle(getTitle());
		builder.setPositiveButton(android.R.string.ok,
			new DontShowAgainOKListener(this, alert_view,
				SettingsActivity.KEY_PREF_SHOW_SENSORS_ALERT));
		builder.setMessage(R.string.message_no_magnetic_sensors);
		builder.setView(alert_view);
		builder.show();
	}

	@Override
	public void onResume() {
		super.onResume();

		this.monitor = new Monitor(this);
		this.monitor.setTarget(this.target != null ? this.target.getLocation() : null);
		this.monitor.execute((Object[])null);

		View compass = findViewById(R.id.compass_arrow);
		ViewTreeObserver observer = compass.getViewTreeObserver();
		if (observer.isAlive()) {
			observer.addOnGlobalLayoutListener(this);
		}

		this.device_wrapper.setupAccelerometer(this.monitor);
		this.device_wrapper.setupMagneticSensor(this.monitor);
		this.device_wrapper.setupGPS(this.monitor);

		this.getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

		maybeStartDistressCallService();
	}

	@Override
	public void onPause() {
		super.onPause();

		this.device_wrapper.releaseGPS(this.monitor);
		this.device_wrapper.releaseMagneticSensor(this.monitor);
		this.device_wrapper.releaseAccelerometer(this.monitor);

		if (this.monitor != null) {
			this.monitor.die();
			this.monitor = null;
		}
	}

	protected void onActivityResult(int request_code, int result_code, Intent data) {
		if (result_code != RESULT_OK) {
			return;
		}

		switch (request_code) {
		case REQUEST_BOOKMARK_RESULT:
			POI poi = (POI)data.getSerializableExtra(BookmarksActivity.Extras.POI);
			assert(poi != null);

			updateTarget(poi);
			break;
		}
	}

	private void rotateCompass(float from, float to, long duration, int repeat_count) {
		float distance_ccw = Math.abs(to - from);
		float distance_cw = 360 - Math.abs(to - from);

		// always rotate compass using the shortest distance
		if (distance_cw < distance_ccw) {
			to = from + (to < from ? +distance_cw : -distance_cw);
		}

		View compass = findViewById(R.id.compass_arrow);

		float around_x = compass.getWidth() / 2;
		float around_y = compass.getHeight() / 2;

		RotateAnimation rotation = new RotateAnimation(from, to, around_x, around_y);
		rotation.setDuration(duration);
		rotation.setRepeatCount(repeat_count);
		rotation.setFillAfter(true);

		compass.startAnimation(rotation);
	}

	private void updateDirection(float azimuth) {
		// round azimuth to direction
		// otherwise constant compass rotation will be bad for a battery
		float direction =  Utils.azimuthToDirection(this.screen.getCurrent(), Config.COMPASS_DIRECTIONS);
		float new_direction = Utils.azimuthToDirection(this.screen.filter(azimuth), Config.COMPASS_DIRECTIONS);

		if (Math.abs(direction - new_direction) < 0.01) {
			return;
		}

		rotateCompass(direction, new_direction, Config.COMPASS_ANIMATION_DURATION, 0);
	}

	private void showDistance(float meters) {
		TextView distance_text = (TextView)findViewById(R.id.distance_text);
		distance_text.setText(UnitsFormatting.formatDistance(this, meters));
	}

	private void updateDistance(float meters) {
		if (this.target != null && this.location != null) {
			showDistance(meters);
		}
	}

	private void updateLocation(final Location location) {
		setLocation(location);
		updateSendLocationMenuItem(this.menu);
	}

	private void showTarget(final POI target) {
		TextView target_text = (TextView)findViewById(R.id.target_text);
		target_text.setText(target.title);
	}

	private void updateTarget(final POI target) {
		// trigger this message only on target update
		// since app can work as compass w/o gps
		if (!device_wrapper.isGPSAvailable()) {
			AlertDialog.Builder builder = new AlertDialog.Builder(this);
			builder.setMessage(R.string.message_gps_not_available);
			builder.setTitle(R.string.app_name);
			builder.setPositiveButton(android.R.string.ok, null);
			builder.create().show();

			return;
		}

		setTarget(target);

		if (this.target != null) {
			showTarget(target);
		}
	}

	private void setLocation(Location location) {
		this.location = location;
	}

	private void setTarget(POI target) {
		this.target = target;
	}

	private void maybeShowCalibrationAlert() {
		final SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);
		boolean show_calibration_alert = shared_pref.getBoolean(SettingsActivity.KEY_PREF_SHOW_CALIBRATION_ALERT, true);

		if (!show_calibration_alert) {
			return;
		}

		LayoutInflater layout_inflater = LayoutInflater.from(this);
		final View alert_view = layout_inflater.inflate(R.layout.layout_showonce_alert, null);

		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setTitle(getTitle());
		builder.setPositiveButton(android.R.string.ok,
			new DontShowAgainOKListener(this, alert_view,
				SettingsActivity.KEY_PREF_SHOW_CALIBRATION_ALERT));
		builder.setMessage(R.string.message_magnetic_calibration);
		builder.setView(alert_view);
		builder.show();
	}

	private void updateMagneticAccuracy(boolean unreliable, boolean direction_magnetic) {
		// only show alert if direction is magnetic
		if (unreliable && direction_magnetic) {
			maybeShowCalibrationAlert();
		}
	}

	private void maybeShowCloseToMagnetAlert() {
		final SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);
		boolean show_range_alert = shared_pref.getBoolean(SettingsActivity.KEY_PREF_SHOW_MAGNETIC_RANGE_ALERT, true);

		if (!show_range_alert) {
			return;
		}

		LayoutInflater layout_inflater = LayoutInflater.from(this);
		final View alert_view = layout_inflater.inflate(R.layout.layout_showonce_alert, null);

		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setTitle(getTitle());
		builder.setPositiveButton(android.R.string.ok,
			new DontShowAgainOKListener(this, alert_view,
				SettingsActivity.KEY_PREF_SHOW_MAGNETIC_RANGE_ALERT));
		builder.setMessage(R.string.message_magnetic_not_in_range);
		builder.setView(alert_view);
		builder.show();
	}

	private void updateMagneticInRange(boolean in_range, boolean direction_magnetic) {
		// only show alert if direction is magnetic
		if (!in_range && direction_magnetic) {
			maybeShowCloseToMagnetAlert();
		}
	}

	private int selectAlertMessage(final Monitor.Entry entry) {
		int preferred_message = -1;

		// in order of priority
		if (DistressCallService.indistress) {
			preferred_message = R.string.alert_distress_call;
		}
		else if (entry.direction_is_magnetic && !entry.magnetic_in_range) {
			preferred_message = R.string.alert_close_to_magnet;
		}
		else if (entry.direction_is_magnetic && entry.magnetic_accuracy_low) {
			preferred_message = R.string.alert_need_calibration;
		}
		// for Stas
		else {
			SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);
			boolean show_waiting_location_alert = shared_pref.getBoolean(SettingsActivity.KEY_PREF_SHOW_WAITING_ALERT, true); // for Max

			if (show_waiting_location_alert && entry.location == null) {
				preferred_message = R.string.alert_waiting_location;
			}
		}

		return preferred_message;
	}

	private void updateAlertMessage(Monitor.Entry entry) {
		TextView alert_text = (TextView)findViewById(R.id.alert_text);
		assert(alert_text != null);

		int message = selectAlertMessage(entry);

		// no preferred message - set empty message
		if (message < 0) {
			alert_text.setText(null);
			return;
		}

		// if there is some message already - reset to make it blink
		CharSequence view_text = alert_text.getText();
		if (view_text != null && view_text.length() > 0) {
			alert_text.setText(null);
			return;
		}

		alert_text.setText(message);
	}

	private void updateTitle(boolean magnetic, boolean from_gps, boolean indistress) {
		if (indistress) {
			setTitle(R.string.title_distress_call);
		}
		else if (magnetic) {
			setTitle(R.string.title_magnetic_compass);
		}
		else if (from_gps) {
			setTitle(R.string.title_gps_compass);
		}
		else {
			setTitle(R.string.title_magnetic_compass_doesnt_work);
		}
	}

	private void updateVisibility(boolean indistress) {
		View target_text = findViewById(R.id.target_text);
		View distance_text = findViewById(R.id.distance_text);
		View compass_bg = findViewById(R.id.compass_bg);
		View compass_arrow = findViewById(R.id.compass_arrow);

		View cancel_button = findViewById(R.id.distress_cancel_button);

		int visibility = (indistress ? View.INVISIBLE : View.VISIBLE);
		int indistress_visibility = (indistress ? View.VISIBLE : View.INVISIBLE);

		target_text.setVisibility(visibility);
		distance_text.setVisibility(visibility);
		compass_bg.setVisibility(visibility);
		compass_arrow.setVisibility(visibility);

		cancel_button.setVisibility(indistress_visibility);
	}

	private void updateGraphics(boolean magnetic, boolean from_gps, boolean indistress) {
		// this is bitmask representing changes need to be made on graphics
		// if it required: bg change on gps direction, visuals change on distress, etc
		int graphics_state = 0
			+ (indistress ? 1 : 0) * 4
			+ (magnetic ? 1 : 0) * 2
			+ (from_gps ? 1 : 0);

		if (this.graphics_state == graphics_state) {
			return;
		}

		this.graphics_state = graphics_state;

		updateVisibility(indistress);

		if (indistress) {
			View compass_arrow = findViewById(R.id.compass_arrow);
			View cancel_button = findViewById(R.id.distress_cancel_button);

			compass_arrow.clearAnimation();
			this.screen.reset(); // to complement clearAnimation()

			cancel_button.setOnClickListener(new View.OnClickListener() {
				@Override
				public void onClick(View v) {
					Intent cancel_intent = new Intent(DistressCallService.ACTION_CANCEL);
					sendBroadcast(cancel_intent);
				}
			});

			return;
		}

		setupImaginery(from_gps);
		this.graphics_state = graphics_state;
	}

	public void updateStatus(Monitor.Entry entry) {
		if (!this.hasWindowFocus()) { // update UI if only it has focus
			return;
		}

		boolean indistress = DistressCallService.indistress;

		updateAlertMessage(entry);

		updateTitle(entry.direction_is_magnetic, entry.direction_from_gps, indistress);
		updateGraphics(entry.direction_is_magnetic, entry.direction_from_gps, indistress);

		updateLocation(entry.location);

		if (indistress) { // don't update compass if in distress mode
			return;
		}

		updateDirection(entry.direction);
		updateDistance(entry.distance);

		updateMagneticAccuracy(entry.magnetic_accuracy_low, entry.direction_is_magnetic);
		updateMagneticInRange(entry.magnetic_in_range, entry.direction_is_magnetic);
	}

	private void setupImaginery(boolean from_gps) {
		ImageView compass_arrow = (ImageView)findViewById(R.id.compass_arrow);
		ImageView compass_bg = (ImageView)findViewById(R.id.compass_bg);

		compass_arrow.setImageDrawable(from_gps
			? this.theme.getGPSCompassArrow(compass_arrow.getWidth(), compass_arrow.getHeight())
			: this.theme.getMagneticCompassArrow(compass_arrow.getWidth(), compass_arrow.getHeight()));

		compass_bg.setImageDrawable(from_gps
			? this.theme.getGPSCompassBG(compass_bg.getWidth(), compass_bg.getHeight())
			: this.theme.getMagneticCompassBG(compass_bg.getWidth(), compass_bg.getHeight()));
	}

	public void onGlobalLayout() {
		SharedPreferences shared_pref = PreferenceManager.getDefaultSharedPreferences(this);

		Theme theme = ThemesFactory.getTheme(this,
				shared_pref.getString(SettingsActivity.KEY_PREF_THEME,
					ThemesFactory.VANILLA));

		if (!DistressCallService.indistress // don't update anything if in distress call
			&& (this.theme != null && !theme.getClass().equals(this.theme.getClass()))) {
			this.theme = theme;
			setupImaginery(false);
		}

		if (this.theme == null) {
			this.theme = theme;
		}

		View compass_bg = findViewById(R.id.compass_arrow);

		// unsubscribe from updates
		ViewTreeObserver observer = compass_bg.getViewTreeObserver();
		if (observer.isAlive()) {
			observer.removeGlobalOnLayoutListener(this);
		}
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.menu_main, menu);
		this.menu = menu;

		return true;
	}

	@Override
	public void onOptionsMenuClosed(Menu menu) {
		this.menu = null;
	}

	private void updateSendLocationMenuItem(Menu menu) {
		if (menu != null) {
			menu.findItem(R.id.menu_send_location).setEnabled(this.location != null);
		}
	}

	@Override
	public boolean onPrepareOptionsMenu(Menu menu) {
		updateSendLocationMenuItem(menu);
		return super.onPrepareOptionsMenu(menu);
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.menu_bookmarks:
		{
			Intent intent = new Intent(MainActivity.this, BookmarksActivity.class);
			if (this.location != null) {
				intent.putExtra(BookmarksActivity.Extras.CURRENT_LOCATION, this.location);
			}

			startActivityForResult(intent, REQUEST_BOOKMARK_RESULT);
		}
			break;

		case R.id.menu_send_location:
			if (this.location != null) {
				ActivityUtils.sendPOI(this, new POI(this.location, null));
			}
			break;

		case R.id.menu_settings:
		{
			Intent intent = new Intent(MainActivity.this, SettingsActivity.class);
			startActivity(intent);
		}
			break;
		}

		return false;
	}
}
